import torch
import torch.nn as nn
import math
import torch.nn.functional as F

from misc.layer import Conv2d, FC
from .efficientnet import EfficientNet
from .bifpn import BIFPN
from .retinahead import RetinaHead
from .module import RegressionModel, ClassificationModel, Anchors, ClipBoxes, BBoxTransform
from torchvision.ops import nms
from .losses import FocalLoss
MODEL_MAP = {
    'efficientdet-d0': 'efficientnet-b0',
    'efficientdet-d1': 'efficientnet-b1',
    'efficientdet-d2': 'efficientnet-b2',
    'efficientdet-d3': 'efficientnet-b3',
    'efficientdet-d4': 'efficientnet-b4',
    'efficientdet-d5': 'efficientnet-b5',
    'efficientdet-d6': 'efficientnet-b6',
    'efficientdet-d7': 'efficientnet-b6',
}


class EfficientDet(nn.Module):
    def __init__(self,
                 num_classes,
                 network='efficientdet-d0',
                 D_bifpn=3,
                 W_bifpn=88,
                 D_class=3,
                 is_training=True,
                 threshold=0.01,
                 iou_threshold=0.5):
        super(EfficientDet, self).__init__()
        self.backbone = EfficientNet.from_pretrained(MODEL_MAP[network])
        self.is_training = is_training
        self.neck = BIFPN(in_channels=self.backbone.get_list_features()[-5:],
                          out_channels=W_bifpn,
                          stack=D_bifpn,
                          num_outs=5)

        self.anchors = Anchors()
        self.regressBoxes = BBoxTransform()
        self.clipBoxes = ClipBoxes()
        self.threshold = threshold
        self.iou_threshold = iou_threshold
        self.de_pred = nn.Sequential(Conv2d(800, 1, 1, same_padding=True, NL='relu'))
                                     #Conv2d(128, 1, 1, same_padding=True, NL='relu'))

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
        self.freeze_bn()
        self.criterion = FocalLoss()

    def forward(self, inputs):
        if self.is_training:
            inputs = inputs
        else:
            inputs = inputs

        x = self.extract_feat(inputs)
        x = self.output_layer(x)
        #import IPython; IPython.embed()
        x = F.interpolate(x, scale_factor = 8)
        return x

    def freeze_bn(self):
        '''Freeze BatchNorm layers.'''
        for layer in self.modules():
            if isinstance(layer, nn.BatchNorm2d):
                layer.eval()

    def extract_feat(self, img):
        """
            Directly extract features from the backbone+neck
        """
        x = self.backbone(img)
        x = self.neck(x[-5:])
        return x

    def output_layer(self, x):
        
        x5 = F.interpolate(x[4], scale_factor=16)
        x4 = F.interpolate(x[3], scale_factor=8)
        x3 = F.interpolate(x[2], scale_factor=4)
        x2 = F.interpolate(x[1], scale_factor=2)
        x1 = x[0]


        xcat = torch.cat([x5,x4,x3,x2,x1],dim=1)
        out = self.de_pred(xcat)
        return out
